import os
import time
import torch
import pickle
import argparse
import numpy as np
from model import *
from utils import Data, split_validation


def init_seed(seed=None):
    if seed is None:
        seed = int(time.time() * 1000 // 1000)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)


parser = argparse.ArgumentParser()
parser.add_argument('--dataset', default='sample', help='dataset name: diginetica/Tmall/RetailRocket/sample')
parser.add_argument('--batchSize', type=int, default=1024, help='input batch size')
parser.add_argument('--hiddenSize', type=int, default=256, help='hidden state size')
parser.add_argument('--lr', type=float, default=0.001, help='learning rate')
parser.add_argument('--lr_dc', type=float, default=0.1, help='learning rate decay rate')
parser.add_argument('--lr_dc_step', type=int, default=3, help='the number of steps after which the learning rate decay')
parser.add_argument('--l2', type=float, default=1e-5, help='l2 penalty')
parser.add_argument('--step', type=int, default=1, help='local session gnn propogation steps')
parser.add_argument('--ratio', type=float, default=0.1, help='ssl ratio')
parser.add_argument('--steps', type=int, default=1, help='global session gnn propogation steps')
parser.add_argument('--recent', type=int, default=20, help='recent number of items')
parser.add_argument('--dev', type=int, default=0, help="cuda device id")
parser.add_argument('--aug', type=float, default=0.3, help="data augment ratio")
parser.add_argument('--k', type=int, default=20, help="top@k")
parser.add_argument('--mu', type=float, default=1, help="the ratio of OSG module")
parser.add_argument('--t', type=float, default=0.2, help="scale rate")
parser.add_argument('--dp', type=float, default=0.1, help="global item dropout ratio")
parser.add_argument('--dp2', type=float, default=0.1, help="global context vector dropout ratio")
parser.add_argument('--dp3', type=float, default=0.1, help="local session dropout ratio")
parser.add_argument('--validation', action='store_true', help='validation')
parser.add_argument('--valid_portion', type=float, default=0.1,
                    help='split the portion of training set as validation set')
opt = parser.parse_args()
print(opt)

epochs = 30
patience = 4
os.environ['CUDA_VISIBLE_DEVICES'] = str(opt.dev)


# os.environ['CUDA_LAUNCH_BLOCKING'] = "1"


def main():
    # init_seed(2021)

    train_data = pickle.load(open('./datasets/' + opt.dataset + '/train.txt', 'rb'))
    if opt.validation:
        train_data, valid_data = split_validation(train_data, opt.valid_portion)
        test_data = valid_data
    else:
        test_data = pickle.load(open('./datasets/' + opt.dataset + '/test.txt', 'rb'))

    train_data = Data(train_data, opt, shuffle=True)
    test_data = Data(test_data, opt, shuffle=True)

    if opt.dataset == 'diginetica':
        n_node = 43098
    elif opt.dataset == 'Tmall':
        n_node = 40728
    elif opt.dataset == 'RetailRocket':
        n_node = 36969
    else:
        n_node = 310

    model = trans_to_cuda(GlobalSessionGraph(opt, n_node))

    start = time.time()
    best_result = [0, 0, 0, 0]
    best_epoch = [0, 0, 0, 0]
    bad_counter = 0
    for epoch in range(epochs):
        print('-------------------------------------------------------')
        print('epoch: ', epoch)
        hit, mrr, hit2, mrr2 = train_test(model, train_data, test_data, opt.ratio, epoch, opt.k)
        flag = 0
        if hit >= best_result[0]:
            best_result[0] = hit
            best_epoch[0] = epoch
            flag = 1
        if mrr >= best_result[1]:
            best_result[1] = mrr
            best_epoch[1] = epoch
            flag = 1
        if hit2 >= best_result[2]:
            best_result[2] = hit2
            best_epoch[2] = epoch
            flag = 1
        if mrr2 >= best_result[3]:
            best_result[3] = mrr2
            best_epoch[3] = epoch
            flag = 1

        print('Best Result(top@20):')
        print('\tHR@:\t%.4f\tMMR@:\t%.4f\tEpoch:\t%d,\t%d' % (
            best_result[0], best_result[1], best_epoch[0], best_epoch[1]))
        print('Best Result(top@10):')
        print('\tHR@:\t%.4f\tMMR@:\t%.4f\tEpoch:\t%d,\t%d' % (
            best_result[2], best_result[3], best_epoch[2], best_epoch[3]))
        bad_counter += 1 - flag
        if bad_counter >= patience:
            break
    print('-------------------------------------------------------')
    end = time.time()
    print("Run time: %f s" % (end - start))


if __name__ == '__main__':
    main()
