import math
import torch
import datetime
import numpy as np
from torch import nn
from torch.nn import Module, Parameter
import torch.nn.functional as F
from gnn import OngoingSessionGraph


class GNN(Module):
    def __init__(self, hidden_size):
        super(GNN, self).__init__()
        self.hidden_size = hidden_size
        self.input_size = hidden_size * 2
        self.gate_size = 3 * hidden_size
        self.w_ih = Parameter(torch.Tensor(self.gate_size, self.input_size))
        self.w_hh = Parameter(torch.Tensor(self.gate_size, self.hidden_size))
        self.b_ih = Parameter(torch.Tensor(self.gate_size))
        self.b_hh = Parameter(torch.Tensor(self.gate_size))
        self.b_iah = Parameter(torch.Tensor(self.hidden_size))
        self.b_oah = Parameter(torch.Tensor(self.hidden_size))

        self.linear_edge_in = nn.Linear(self.hidden_size, self.hidden_size, bias=True)
        self.linear_edge_out = nn.Linear(self.hidden_size, self.hidden_size, bias=True)
        self.linear_edge_f = nn.Linear(self.hidden_size, self.hidden_size, bias=True)

    def GNNCell(self, A, hidden):
        input_in = torch.matmul(A[:, :A.shape[0]], self.linear_edge_in(hidden)) + self.b_iah
        input_out = torch.matmul(A[:, A.shape[0]: 2 * A.shape[0]], self.linear_edge_out(hidden)) + self.b_oah
        inputs = torch.cat([input_in, input_out], -1)
        gi = F.linear(inputs, self.w_ih, self.b_ih)
        gh = F.linear(hidden, self.w_hh, self.b_hh)
        i_r, i_i, i_n = gi.chunk(3, 1)
        h_r, h_i, h_n = gh.chunk(3, 1)
        resetgate = torch.sigmoid(i_r + h_r)
        inputgate = torch.sigmoid(i_i + h_i)
        newgate = torch.tanh(i_n + resetgate * h_n)
        hy = newgate + inputgate * (hidden - newgate)
        return hy

    def forward(self, A, hidden):
        hidden = self.GNNCell(A, hidden)
        return hidden


class GlobalSessionGraph(Module):
    def __init__(self, opt, n_node):
        super(GlobalSessionGraph, self).__init__()
        self.hidden_size = opt.hiddenSize
        self.mu = opt.mu
        self.n_node = n_node
        self.batch_size = opt.batchSize
        self.embedding = nn.Embedding(self.n_node, self.hidden_size)
        self.pos_embedding = nn.Embedding(opt.recent, self.hidden_size)
        self.gnn = OngoingSessionGraph(self.hidden_size, step=opt.step, dp3=opt.dp3)

        self.linear_one = nn.Linear(self.hidden_size, self.hidden_size, bias=True)
        self.linear_two = nn.Linear(self.hidden_size, self.hidden_size, bias=True)
        self.linear_three = nn.Linear(self.hidden_size, 1, bias=False)
        self.linear_transform = nn.Linear(self.hidden_size * 2, self.hidden_size, bias=True)

        self.srgnn = GNN(self.hidden_size)
        self.steps = opt.steps
        self.highwayitem = nn.Linear(2 * self.hidden_size, self.hidden_size)
        self.highwaycontext = nn.Linear(2 * self.hidden_size, self.hidden_size)
        self.update_item = nn.Linear(2 * self.hidden_size, self.hidden_size, bias=True)
        self.update_context = nn.Linear(2 * self.hidden_size, self.hidden_size, bias=True)

        self.dp = opt.dp
        self.dp2 = opt.dp2
        self.dropout1 = nn.Dropout(self.dp)
        self.dropout2 = nn.Dropout(self.dp)
        self.dropout3 = nn.Dropout(self.dp2)
        self.dropout4 = nn.Dropout(self.dp2)
        self.T = opt.t
        self.d = trans_to_cuda(torch.sqrt(torch.tensor(self.hidden_size, dtype=torch.float)))

        self.loss_function = nn.CrossEntropyLoss()
        self.optimizer = torch.optim.Adam(self.parameters(), lr=opt.lr, weight_decay=opt.l2)
        self.scheduler = torch.optim.lr_scheduler.StepLR(self.optimizer, step_size=opt.lr_dc_step, gamma=opt.lr_dc)
        self.reset_parameters()

    def reset_parameters(self):
        stdv = 1.0 / math.sqrt(self.hidden_size)
        for weight in self.parameters():
            weight.data.uniform_(-stdv, stdv)

    def mask_correlated_samples(self, x):
        mask = torch.ones((2 * x, 2 * x), dtype=torch.bool)
        mask = mask.fill_diagonal_(0)
        for i in range(x):
            mask[i, x + i] = 0
            mask[x + i, i] = 0
        return mask

    def info_nce(self, origin_context, aug_context):
        origin_context = F.normalize(origin_context, p=2, dim=1)
        aug_context = F.normalize(aug_context, p=2, dim=1)
        N = 2 * origin_context.shape[0]
        context = torch.cat((origin_context, aug_context), dim=0)
        sim = torch.mm(context, context.T) / self.T
        sim_o_a = torch.diag(sim, origin_context.shape[0])
        sim_a_o = torch.diag(sim, -origin_context.shape[0])

        positive_samples = torch.cat((sim_o_a, sim_a_o), dim=0).reshape(N, 1)
        mask = self.mask_correlated_samples(origin_context.shape[0])

        negative_samples = sim[mask].reshape(N, -1)
        labels = torch.zeros(N).to(positive_samples.device).long()
        logits = torch.cat((positive_samples, negative_samples), dim=1)
        sslloss = self.loss_function(logits, labels)
        return sslloss

    def compute_scores(self, seq_hidden, mask):
        bs, item_num = seq_hidden.shape[0], seq_hidden.shape[1]
        index = torch.arange(item_num).unsqueeze(0)
        pos_index = index.repeat(bs, 1).view(bs, item_num)
        pos_index = trans_to_cuda(torch.Tensor(pos_index.float()).long())
        pos_hidden = self.pos_embedding(pos_index)
        seq_hidden = seq_hidden + pos_hidden

        ht = seq_hidden[torch.arange(mask.shape[0]).long(), torch.sum(mask, 1) - 1]
        q1 = self.linear_one(ht).view(ht.shape[0], 1, ht.shape[1])
        q2 = self.linear_two(seq_hidden)
        alpha = self.linear_three(torch.sigmoid(q1 + q2))
        res = torch.sum(alpha * seq_hidden * mask.view(mask.shape[0], -1, 1).float(), 1)
        res = self.linear_transform(torch.cat([res, ht], 1))
        return res

    def update_item_with_context(self, context_emb, item_embeddings, item_belong_sess):
        sess_emb = context_emb[item_belong_sess]
        beta = torch.bmm(item_embeddings.unsqueeze(1), sess_emb.permute(0, 2, 1)) / self.d
        beta = beta.squeeze(1)
        beta = torch.sigmoid(beta)
        mask = item_belong_sess.float()
        mask[mask < 1.0e-5] = 0
        mask[mask > 1.0e-5] = 1
        beta2 = beta * mask
        mid = torch.bmm(beta2.unsqueeze(1), sess_emb).squeeze(1)
        alpha = torch.cat((item_embeddings, mid), dim=-1)
        alpha = trans_to_cuda(self.update_item)(alpha)
        alpha = torch.sigmoid(alpha)
        res = alpha * item_embeddings + (1 - alpha) * mid
        return res

    def update_context_with_item(self, context_emb, item_embeddings, sess_has_item, mask):
        items_emb = item_embeddings[sess_has_item]
        beta = torch.bmm(context_emb.unsqueeze(1), items_emb.permute(0, 2, 1)) / self.d
        beta = beta.squeeze(1)
        q = torch.max(beta, dim=-1, keepdim=True)
        beta = beta - q.values
        beta = torch.exp(beta)
        mask = mask.float()
        mask = torch.cat((trans_to_cuda(torch.zeros(1, beta.shape[1])), mask), dim=0)
        beta = beta * mask
        beta /= torch.sum(beta, -1).unsqueeze(-1) + 1e-24
        mid = torch.bmm(beta.unsqueeze(1), items_emb).squeeze(1)
        alpha = torch.cat((context_emb, mid), dim=-1)
        alpha = trans_to_cuda(self.update_context)(alpha)
        alpha = torch.sigmoid(alpha)
        res = alpha * context_emb + (1 - alpha) * mid
        return res

    def ave_pooling(self, hidden, graph_mask):
        length = torch.sum(graph_mask, 1)
        hidden = hidden * graph_mask.unsqueeze(-1).float()
        output = torch.sum(hidden, 1) / length.unsqueeze(-1).float()
        return output

    def forward(self, inputs, A, adjacency, adjacency2, graph_mask, all_items, sess_has_item, alias_inputs,
                item_belong_sess, mask, tar):
        hidden = self.embedding(inputs)
        hidden_new = self.gnn(A, hidden)
        get = lambda i: hidden_new[i][alias_inputs[i]]
        seq_hidden = torch.stack([get(i) for i in torch.arange(len(alias_inputs)).long()])
        all_items_emb = self.embedding(all_items)
        context_embs = trans_to_cuda(torch.zeros((sess_has_item.shape[0], self.hidden_size)))
        get2 = lambda i: hidden[i][alias_inputs[i]]
        seq_hidden2 = torch.stack([get2(i) for i in torch.arange(len(alias_inputs)).long()])
        context_embs[1:] = self.compute_scores(seq_hidden2, mask)
        # context_embs[1:] = self.ave_pooling(hidden_new, graph_mask)

        item_embeddings_i0 = all_items_emb
        context_emb_i0 = context_embs
        item_emb = all_items_emb
        context_emb = context_embs
        #
        context_emb_i02 = context_embs
        item_emb2 = all_items_emb
        context_emb2 = context_embs
        #
        for i in range(self.steps):
            item_emb = self.srgnn(adjacency, item_emb)
            item_emb = self.dropout1(item_emb)
            item_emb = self.update_item_with_context(context_emb, item_emb, item_belong_sess)
            context_emb = self.update_context_with_item(context_emb, item_emb, sess_has_item, mask)
            context_emb = self.dropout3(context_emb)
            #
            item_emb2 = self.srgnn(adjacency2, item_emb2)
            item_emb2 = self.dropout2(item_emb2)
            item_emb2 = self.update_item_with_context(context_emb2, item_emb2, item_belong_sess)
            context_emb2 = self.update_context_with_item(context_emb2, item_emb2, sess_has_item, mask)
            context_emb2 = self.dropout4(context_emb2)

        concat_item = torch.cat((item_embeddings_i0, item_emb), dim=-1)
        weight_item = torch.sigmoid(self.highwayitem(concat_item))
        item_emb = weight_item * item_embeddings_i0 + (1 - weight_item) * item_emb

        concat_context = torch.cat((context_emb_i0, context_emb), dim=-1)
        weight_context = torch.sigmoid(self.highwaycontext(concat_context))
        context_emb = weight_context * context_emb_i0 + (1 - weight_context) * context_emb

        concat_context2 = torch.cat((context_emb_i02, context_emb2), dim=-1)
        weight_context2 = torch.sigmoid(self.highwaycontext(concat_context2))
        context_emb2 = weight_context2 * context_emb_i02 + (1 - weight_context2) * context_emb2

        sslloss = self.info_nce(context_emb[1:, :], context_emb2[1:, :])
        seq_global = item_emb[sess_has_item[1:, :]]
        seq_global = seq_global + self.mu * seq_hidden
        sess_emb_i = self.compute_scores(seq_global, mask)
        item_embeddings = self.embedding.weight[1:]
        sess_emb_i = F.normalize(sess_emb_i, p=2, dim=-1)
        item_embeddings = F.normalize(item_embeddings, p=2, dim=-1)
        scores_item = torch.mm(12 * sess_emb_i, torch.transpose(item_embeddings, 1, 0))
        loss_item = self.loss_function(scores_item, tar - 1)
        # sslloss = trans_to_cuda(torch.tensor(0))
        return loss_item, scores_item, sslloss


def trans_to_cuda(variable):
    if torch.cuda.is_available():
        return variable.cuda()
    else:
        return variable


def trans_to_cpu(variable):
    if torch.cuda.is_available():
        return variable.cpu()
    else:
        return variable


def forward(model, i, data):
    alias_inputs, A, adjacency, adjacency2, mask, items, targets, sess_has_item, item_belong_sess, all_items, session_len = data.get_slice_gnn(
        i)
    alias_inputs = trans_to_cuda(torch.Tensor(alias_inputs).long())
    items = trans_to_cuda(torch.Tensor(items).long())
    A = trans_to_cuda(torch.Tensor(np.array(A)).float())
    adjacency = trans_to_cuda(torch.Tensor(adjacency).float())
    adjacency2 = trans_to_cuda(torch.Tensor(adjacency2).float())
    sess_has_item = trans_to_cuda(torch.Tensor(sess_has_item).long())
    item_belong_sess = trans_to_cuda(torch.Tensor(item_belong_sess).long())
    targets = trans_to_cuda(torch.Tensor(targets).long())
    all_items = trans_to_cuda(torch.Tensor(all_items).long())
    mask = trans_to_cuda(torch.Tensor(mask).long())
    graph_mask = torch.sign(items)
    loss_item, scores_item, sslloss = model(items, A, adjacency, adjacency2, graph_mask, all_items, sess_has_item,
                                            alias_inputs, item_belong_sess, mask, targets)
    return targets, scores_item, loss_item, sslloss


def train_test(model, train_data, test_data, ratio, epoch, k):
    print('start training: ', datetime.datetime.now())
    model.train()
    total_loss = 0.0
    slices = train_data.generate_batch(model.batch_size)
    for i, j in zip(slices, np.arange(len(slices))):
        model.optimizer.zero_grad()
        targets, scores_item, loss_item, sslloss = forward(model, i, train_data)
        loss = loss_item + ratio * sslloss
        loss.backward()
        model.optimizer.step()
        total_loss += loss_item.cpu().item()
        if j % int(len(slices) / 5 + 1) == 0:
            print('[%d/%d] Loss: %.4f SSL: %.4f' % (j, len(slices), loss_item.item(), sslloss.item()))
    print('\tLoss:\t%.3f' % total_loss)
    model.scheduler.step()

    print('start predicting: ', datetime.datetime.now())
    model.eval()
    hit, mrr, hit2, mrr2 = [], [], [], []
    slices = test_data.generate_batch(model.batch_size)
    for i, j in zip(slices, np.arange(len(slices))):
        targets, scores, loss_item, sslloss = forward(model, i, test_data)
        targets = targets.detach().cpu().numpy()
        sub_scores = scores.topk(20)[1]
        sub_scores2 = scores.topk(10)[1]
        sub_scores = trans_to_cpu(sub_scores).detach().numpy()
        sub_scores2 = trans_to_cpu(sub_scores2).detach().numpy()
        for score, target in zip(sub_scores, targets):
            hit.append(np.isin(target - 1, score))
            if len(np.where(score == target - 1)[0]) == 0:
                mrr.append(0)
            else:
                mrr.append(1 / (np.where(score == target - 1)[0][0] + 1))
        for score, target in zip(sub_scores2, targets):
            hit2.append(np.isin(target - 1, score))
            if len(np.where(score == target - 1)[0]) == 0:
                mrr2.append(0)
            else:
                mrr2.append(1 / (np.where(score == target - 1)[0][0] + 1))
    hit = np.mean(hit) * 100
    mrr = np.mean(mrr) * 100
    hit2 = np.mean(hit2) * 100
    mrr2 = np.mean(mrr2) * 100
    return hit, mrr, hit2, mrr2
